@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2007 OpenedHand Ltd.
+ * Copyright (C) 2007, 2008 OpenedHand Ltd.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
  *
@@ -146,6 +146,57 @@ gupnp_root_device_get_property (GObject    *object,
         }
 }
 
+/* Adds @udn::@resource_type to @group, together with all earlier versions
+ * of @resource_type.
+ *
+ * Note: @resource_type is modified in this function. */
+static void
+add_resource_with_earlier_versions (GSSDPResourceGroup *group,
+                                    const char         *udn,
+                                    const char         *resource_type,
+                                    const char         *location)
+{
+        char *colon, *usn;
+        int version, i;
+
+        /* Add specified version */
+        usn = g_strdup_printf ("%s::%s", (char *) udn, (char *) resource_type);
+        gssdp_resource_group_add_resource_simple (group,
+                                                  (const char *) resource_type,
+                                                  usn,
+                                                  location);
+        g_free (usn);
+
+        /* Try to parse version */
+        colon = strrchr (resource_type, ':');
+        if (G_UNLIKELY (*colon == 0))
+                return;
+
+        colon += 1;
+        if (G_UNLIKELY (*colon == 0))
+                return;
+
+        version = atoi (colon);
+        
+        /* Add earlier versions */
+        *colon = 0;
+
+        for (i = 1; i < version; i++) {
+                char *type;
+
+                type = g_strdup_printf ("%s%d", (char *) resource_type, i);
+                usn = g_strdup_printf ("%s::%s", (char *) udn, (char *) type);
+
+                gssdp_resource_group_add_resource_simple (group,
+                                                          type,
+                                                          usn,
+                                                          location);
+
+                g_free (usn);
+                g_free (type);
+        }
+}
+
 static void
 fill_resource_group (xmlNode            *element,
                      const char         *location,
@@ -153,7 +204,6 @@ fill_resource_group (xmlNode            *element,
 {
         xmlNode *child;
         xmlChar *udn, *device_type;
-        char *usn;
 
         /* Add device */
         udn = xml_util_get_child_element_content (element, "UDN");
@@ -176,12 +226,10 @@ fill_resource_group (xmlNode            *element,
                                                   (const char *) udn,
                                                   location);
 
-        usn = g_strdup_printf ("%s::%s", (char *) udn, (char *) device_type);
-        gssdp_resource_group_add_resource_simple (group,
-                                                  (const char *) device_type,
-                                                  (const char *) usn,
-                                                  location);
-        g_free (usn);
+        add_resource_with_earlier_versions (group,
+                                            (const char *) udn,
+                                            (const char *) device_type,
+                                            location);
 
         xmlFree (device_type);
 
@@ -201,15 +249,11 @@ fill_resource_group (xmlNode            *element,
                         if (!service_type)
                                 continue;
 
-                        usn = g_strdup_printf ("%s::%s",
-                                               (char *) udn,
-                                               (char *) service_type);
-                        gssdp_resource_group_add_resource_simple
-                                                (group,
-                                                 (const char *) service_type,
-                                                 (const char *) usn,
-                                                 location);
-                        g_free (usn);
+                        add_resource_with_earlier_versions
+                                (group,
+                                 (const char *) udn,
+                                 (const char *) service_type,
+                                 location);
 
                         xmlFree (service_type);
                 }
@@ -243,7 +287,7 @@ gupnp_root_device_constructor (GType                  type,
         int i;
         xmlDoc *description_doc;
         xmlNode *root_element, *element;
-        SoupUri *url_base;
+        SoupURI *url_base;
 
         /* Get 'description-doc' property value */
         description_doc = NULL;
@@ -437,8 +481,40 @@ gupnp_root_device_new (GUPnPContext *context,
                        xmlDoc       *description_doc,
                        const char   *relative_location)
 {
+        GUPnPResourceFactory *factory;
+
+        factory = gupnp_resource_factory_get_default ();
+
+        return gupnp_root_device_new_full (context,
+                                           factory,
+                                           description_doc,
+                                           relative_location);
+}
+
+/**
+ * gupnp_root_device_new_full
+ * @context: A #GUPnPContext
+ * @factory: A #GUPnPResourceFactory
+ * @description_doc: Pointer to the device description document
+ * @relative_location: Location to use for this device, relative to the
+ * HTTP root
+ *
+ * Create a new #GUPnPRootDevice.
+ *
+ * Return value: A new #GUPnPRootDevice object.
+ **/
+GUPnPRootDevice *
+gupnp_root_device_new_full (GUPnPContext         *context,
+                            GUPnPResourceFactory *factory,
+                            xmlDoc               *description_doc,
+                            const char           *relative_location)
+{
+        g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
+        g_return_val_if_fail (GUPNP_IS_RESOURCE_FACTORY (factory), NULL);
+
         return g_object_new (GUPNP_TYPE_ROOT_DEVICE,
                              "context", context,
+                             "resource-factory", factory,
                              "root-device", NULL,
                              "description-doc", description_doc,
                              "relative-location", relative_location,
@@ -469,7 +545,9 @@ gupnp_root_device_set_available (GUPnPRootDevice *root_device,
  * gupnp_root_device_get_available
  * @root_device: A #GUPnPRootDevice
  *
- * Return value: TRUE if @root_device is available.
+ * Get whether or not @root_device is available (announcing its presence).
+ *
+ * Return value: %TRUE if @root_device is available, %FALSE otherwise.
  **/
 gboolean
 gupnp_root_device_get_available (GUPnPRootDevice *root_device)
@@ -483,6 +561,8 @@ gupnp_root_device_get_available (GUPnPRootDevice *root_device)
  * gupnp_root_device_get_relative_location
  * @root_device: A #GUPnPRootDevice
  *
+ * Get the relative location of @root_device.
+ *
  * Return value: The relative location of @root_device.
  **/
 const char *
